package es.kauron.jstudy.controller;

import es.kauron.jstudy.model.AppPrefs;
import es.kauron.jstudy.model.TestItem;
import javafx.beans.binding.Bindings;
import javafx.beans.property.*;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.collections.transformation.FilteredList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.control.*;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseEvent;
import javafx.stage.FileChooser;

import java.io.File;
import java.net.URL;
import java.util.*;

public class TableController implements Initializable {
    @FXML
    private TableView<TestItem> table;
    @FXML
    private TextField newQuestionField, newAnswerField, searchField;
    @FXML
    private Button addButton;
    @FXML
    private TableColumn<TestItem, String> answerCol, questionCol;

    private FilteredList<TestItem> filtered;
    private ObservableList<TestItem> data;
    private Controller parent;
    private File file;
    StringProperty name;
    final BooleanProperty saved = new SimpleBooleanProperty();
    private final ObjectProperty<TestItem> editing = new SimpleObjectProperty<>(null);

    @Override
    public void initialize(URL url, ResourceBundle resourceBundle) {
        addButton.textProperty().bind(Bindings.when(addButton.disableProperty())
                .then("Question repeated")
                .otherwise(Bindings.when(editing.isNull())
                        .then("_Add item")
                        .otherwise("_Save item")));
        addButton.getTooltip().textProperty().bind(Bindings.when(editing.isNull())
                .then("Add an entry to the table")
                .otherwise("Save changes in the selected entry"));
        // Add context menu to Table
        MenuItem menuEdit = new MenuItem("_Edit");
        menuEdit.setOnAction(this::onEditAction);
        MenuItem menuDelete = new MenuItem("_Delete");
        menuDelete.setOnAction(this::onDeleteAction);
        MenuItem menuDup = new MenuItem("D_uplicate");
        menuDup.setOnAction(this::onDuplicateAction);
        MenuItem menuSwap = new MenuItem("_Swap");
        menuSwap.setOnAction(this::onSwapAction);
        MenuItem menuTest = new MenuItem("_Test selected");
        menuTest.setOnAction(this::onTestSelectionAction);
        table.setContextMenu(new ContextMenu(menuEdit, menuDelete, menuDup, menuSwap, new SeparatorMenuItem(), menuTest));

        table.getSelectionModel().setSelectionMode(SelectionMode.MULTIPLE);
        answerCol.setCellValueFactory(e -> e.getValue().answerProperty());
        questionCol.setCellValueFactory(e -> e.getValue().questionProperty());
        table.getSelectionModel().getSelectedIndices().addListener(
                (ListChangeListener<? super Integer>) obs -> menuEdit.setDisable(obs.getList().size() != 1));
        searchField.textProperty().addListener((obj, o, n) ->
                filtered.setPredicate((item) -> item.getQuestion().contains(n) || item.getAnswer().contains(n)));
        newQuestionField.textProperty().addListener((obj, o, n) -> {
            for (TestItem item : data) {
                if (item == editing.get()) continue;
                if (item.getQuestion().equals(n.trim())) {
                    addButton.setDisable(true);
                    return;
                }
            }
            addButton.setDisable(false);
        });
    }

    void setData(String name, List<TestItem> list, Controller controller, File file) {
        this.name = new SimpleStringProperty(name);
        this.data = FXCollections.observableArrayList(list);
        this.filtered = data.filtered((item) -> true);
        this.parent = controller;
        table.setRowFactory(new DraggableRowFactory<>(data)::generator);
        table.setItems(filtered);
        this.file = file;
        saved.set(file != null);
    }

    List<TestItem> getData() {return new ArrayList<>(data);}
    String getName() {return name.get();}

    protected void onSaveAction(ActionEvent event) {
        while (file == null) {
            FileChooser chooser = new FileChooser();
            chooser.getExtensionFilters().add(new FileChooser.ExtensionFilter("JStudy file", "*.jsdb"));
            if (AppPrefs.lastDir != null) chooser.setInitialDirectory(AppPrefs.lastDir);
            chooser.setInitialFileName(name.get());
            file = chooser.showSaveDialog(table.getScene().getWindow());
            if (!file.getName().endsWith(".jsdb"))
                file = new File(file.getPath() + ".jsdb");
            if (file != null) {
                AppPrefs.lastDir = file.getParentFile();
                name.set(file.getName().substring(0, file.getName().lastIndexOf('.')));
            }
        }
        TestItem.saveTo(file, data);
        saved.set(true);
    }

    @FXML
    protected void onAddAction(ActionEvent event) {
        if (newQuestionField.getText().trim().isEmpty() && newAnswerField.getText().trim().isEmpty()) {
            newQuestionField.requestFocus();
            return;
        }
        if (editing.get() == null) {
            TestItem item = new TestItem(newQuestionField.getText().trim(), newAnswerField.getText().trim());
            data.add(item);
            saved.set(false);
        } else {
            editing.get().answerProperty().set(newAnswerField.getText().trim());
            editing.get().questionProperty().set(newQuestionField.getText().trim());
        }
        onCancelAction(event);
    }

    protected void onEditAction(ActionEvent event) {
        ObservableList<TestItem> list = table.getSelectionModel().getSelectedItems();
        if (list.size() != 1) return;
        editing.set(list.get(0));
        newQuestionField.setText(list.get(0).getQuestion());
        newAnswerField.setText(list.get(0).getAnswer());
        saved.set(false);
        newQuestionField.requestFocus();
    }

    @FXML
    protected void onCancelAction(ActionEvent event) {
        if (editing.get() != null)
            editing.set(null);
        newQuestionField.setText("");
        newAnswerField.setText("");
        newQuestionField.requestFocus();
    }

    protected void onSwapAction(ActionEvent event) {
        if (table.getSelectionModel().getSelectedIndices().size() > 0) saved.set(false);
        for (TestItem item : table.getSelectionModel().getSelectedItems()) {
            String question = item.getQuestion();
            item.questionProperty().setValue(item.getAnswer());
            item.answerProperty().setValue(question);
        }
        table.requestFocus();
    }

    protected void onDuplicateAction(ActionEvent event) {
        if (table.getSelectionModel().getSelectedIndices().size() > 0) saved.set(false);
        for (int i : table.getSelectionModel().getSelectedIndices()) {
            TestItem item = new TestItem(filtered.get(i));
            item.questionProperty().setValue(item.getQuestion() + " 2");
            data.add(item);
        }
        table.requestFocus();
    }

    protected void onDeleteAction(ActionEvent event) {
        if (table.getSelectionModel().getSelectedIndices().size() > 0) saved.set(false);
        data.removeAll(table.getSelectionModel().getSelectedItems());
        table.requestFocus();
    }

    protected void onTestSelectionAction(ActionEvent event) {
        parent.newTest(table.getSelectionModel().getSelectedItems());
    }

    @FXML
    protected void onTestAction(ActionEvent event) {
        parent.newTest(filtered);
    }

    @FXML
    protected void onTableKeyEvent(KeyEvent event) {
        if (event.getCode().equals(KeyCode.DELETE)) {
            onDeleteAction(null);
        }
    }

    private final Timer timer = new Timer();
    private TimerTask timerTask = null;
    @FXML
    protected void onTableMouseClicked(MouseEvent event) {
        if (timerTask == null) {
            timerTask = new TimerTask() {
                @Override
                public void run() {
                    timerTask = null;
                }
            };
            timer.schedule(timerTask, 200);
        } else {
            if (timerTask.cancel())
                onEditAction(null);
            timerTask = null;
        }
    }
}
